/* passivesock.c - passivesock */

#define _POSIX_C_SOURCE 200112L				/*	or #define _POSIX_C_SOURCE 200809L	*/

#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <err.h>
#include <errno.h>

unsigned short	portbase = 0;	/* port base, for non-root servers	*/

/*------------------------------------------------------------------------
 * passivesock - allocate & bind a server socket using TCP or UDP
 *------------------------------------------------------------------------
 */
int
passivesock(const char *service, const char *transport, const int qlen)
/*
 * Arguments:
 *      service   - service associated with the desired port
 *      transport - transport protocol to use ("tcp" or "udp")
 *      qlen      - maximum server request queue length
 */
{
	struct addrinfo *res, *resptr;
	struct addrinfo	hints = { 0 };
	struct sockaddr_in	*sa_in;
	struct sockaddr_in6	*sa_in6;
	int ret;
	int pasvsd, type; /* socket descriptor and socket type */
	int off = 0, on = 1;	

	/* Use transport protocol to choose a socket type */
	if (strcmp(transport, "UDP") == 0)
		type = SOCK_DGRAM;
	else if (strcmp(transport, "TCP") == 0)
		type = SOCK_STREAM;
	else
		errx(1, "passivesock arguments error: unknown transport\n");

	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = type;
	hints.ai_flags = AI_PASSIVE;

	if ((ret = getaddrinfo(NULL, service, &hints, &res)) != 0)
		errx(1, "getaddrinfo: %s\n", gai_strerror(ret));

	for (resptr = res; resptr != NULL; resptr = resptr->ai_next) {
		/* Compute a final portnumber as a function of the standard port and the value of portbase.	*/
		if (portbase > 0){
			if (resptr->ai_family == AF_INET) {
				sa_in = (struct sockaddr_in *)resptr->ai_addr;
				sa_in->sin_port = htons(ntohs((unsigned short)sa_in->sin_port) + portbase);
			}
			else if (resptr->ai_family == AF_INET6) {
				sa_in6 = (struct sockaddr_in6 *)resptr->ai_addr;
				sa_in6->sin6_port = htons(ntohs((unsigned short)sa_in6->sin6_port) + portbase);
			};
		}

		/* Allocate a socket */
		if ((pasvsd = socket(resptr->ai_family, resptr->ai_socktype, resptr->ai_protocol)) < 0)
			ret = errno;
		
		/* Enable local address reuse.  */
		else if (resptr->ai_socktype == SOCK_STREAM && setsockopt(pasvsd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
			ret = errno;
			close(pasvsd);
		}
		
		/* Upgrade to dual stacked socket.  */
		else if (resptr->ai_family == AF_INET6 && setsockopt(pasvsd, IPPROTO_IPV6, 
			IPV6_V6ONLY, &off, sizeof(off)) < 0) {
			ret = errno;
			close(pasvsd);
		}
		
		/* Bind the socket */
		else if (bind(pasvsd, resptr->ai_addr, resptr->ai_addrlen) < 0){
			ret = errno;
			close(pasvsd);
		}
		
		else if (resptr->ai_socktype == SOCK_STREAM && listen(pasvsd, qlen) < 0){
			ret = errno;
			close(pasvsd);
		}
		
		else {
			freeaddrinfo(res);
			return pasvsd;
		}
		
	}
	errno = ret;
	errx(1, "can't allocate & bind %s socket: %s\n", service, strerror(errno));
}

/**************************************************************************
 * Building a shared library(DLLs):                                       *
 * gcc -Wall -pedantic -std=c11 -g -shared -fpic                          *
 *     -o passivesock.dll passivesock.c passiveTCP.c passiveUDP.c         *
 *                                                                        *
 *                                                                        *
 * To link:                                                               *
 * -L ../api -l passivesock                                               *
 *                                                                        *
 * To run:                                                                *
 * cp passivesock.dll ../apps/passivesock.dll                             *
 **************************************************************************/