/* TCPtecho.c - main, TCPtecho, reader, writer, mstime */

#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <err.h>
#include <errno.h>

int TCPtecho(fd_set *, const int, const int, int);
int reader(const int, fd_set *);
int writer(const int, fd_set *);
int connectTCP(const char *, const char *);
long mstime(unsigned long *);

#ifndef MIN
#define MIN(x, y) ((x) < (y) ? (x) : (y))
#endif /* MIN */

#define BUFSIZE 8192	 /* write buffer size		*/
#define CCOUNT 64 * 1024 /* default character count	*/

#define USAGE "usage: TCPtecho [ -c count ] host1 host2...\n"

char *hname[FD_SETSIZE];			/* fd to host name mapping	*/
int rc[FD_SETSIZE], wc[FD_SETSIZE]; /* read/write character counts	*/
char buf[BUFSIZE];					/* read/write data buffer	*/

/*------------------------------------------------------------------------
 * main - concurrent TCP client for ECHO service timing
 *------------------------------------------------------------------------
 */
int main(int argc, char *argv[])
{
	int ccount = CCOUNT;
	int i, hcount, maxfd, fd;
	int one = 1;
	fd_set afds;

	hcount = 0;
	maxfd = -1;
	for (i = 1; i < argc; ++i)
	{
		if (strcmp(argv[i], "-c") == 0)
		{
			if (++i < argc && (ccount = atoi(argv[i])))
				continue;
			errx(0, USAGE);
		}
		/* else, a host */

		fd = connectTCP(argv[i], "echo");
		if (ioctl(fd, FIONBIO, (char *)&one))
			err(1, "main() ioctl can't mark socket nonblocking");
		if (fd > maxfd)
			maxfd = fd;
		hname[fd] = argv[i];
		++hcount;
		FD_SET(fd, &afds);
	}
	TCPtecho(&afds, maxfd + 1, ccount, hcount);
	exit(0);
}

/*------------------------------------------------------------------------
 * TCPtecho - time TCP ECHO requests to multiple servers
 *------------------------------------------------------------------------
 */
int TCPtecho(fd_set *pafds, const int nfds, const int ccount, int hcount)
{
	fd_set rfds, wfds;	 /* read/write fd sets		*/
	fd_set rcfds, wcfds; /* read/write fd sets (copy)	*/
	int fd, i;

	for (i = 0; i < BUFSIZE; ++i) /* echo data	*/
		buf[i] = 'D';
	memcpy(&rcfds, pafds, sizeof(rcfds));
	memcpy(&wcfds, pafds, sizeof(wcfds));
	for (fd = 0; fd < nfds; ++fd)
		rc[fd] = wc[fd] = ccount;

	(void)mstime((unsigned long *)0); /* set the epoch */

	while (hcount)
	{
		memcpy(&rfds, &rcfds, sizeof(rfds));
		memcpy(&wfds, &wcfds, sizeof(wfds));

		if (select(nfds, &rfds, &wfds, (fd_set *)0,
				   (struct timeval *)0) < 0)
			err(1, "TCPtecho() select");
		for (fd = 0; fd < nfds; ++fd)
		{
			if (FD_ISSET(fd, &rfds))
				if (reader(fd, &rcfds) == 0)
					hcount--;
			if (FD_ISSET(fd, &wfds))
				writer(fd, &wcfds);
		}
	}

	return 0;
}

/*------------------------------------------------------------------------
 * reader - handle ECHO reads
 *------------------------------------------------------------------------
 */
int reader(const int fd, fd_set *pfdset)
{
	unsigned long now;
	ssize_t cc;

	cc = read(fd, buf, sizeof(buf));
	if (cc < 0)
		err(1, "reader() read");
	if (cc == 0)
		err(1, "reader() read: premature end of file");
	rc[fd] -= cc;
	if (rc[fd])
		return 1;
	(void)mstime(&now);
	printf("%s: %d ms\n", hname[fd], (int)now);
	(void)close(fd);
	FD_CLR(fd, pfdset);
	return 0;
}

/*------------------------------------------------------------------------
 * writer - handle ECHO writes
 *------------------------------------------------------------------------
 */
int writer(const int fd, fd_set *pfdset)
{
	ssize_t nchars;

	nchars = write(fd, buf, MIN((int)sizeof(buf), wc[fd]));
	if (nchars < 0)
		err(1, "writer() write");
	wc[fd] -= nchars;
	if (wc[fd] == 0)
	{
		(void)shutdown(fd, 1);
		FD_CLR(fd, pfdset);
	}
	return 0;
}

/*------------------------------------------------------------------------
 * mstime - report the number of milliseconds elapsed
 *------------------------------------------------------------------------
 */
long mstime(unsigned long *pms)
{
	static struct timespec epoch;
	struct timespec now;

	if (clock_gettime(CLOCK_REALTIME, &now))
		err(1, "mstime() clock_gettime");
	if (!pms)
	{
		epoch = now;
		return 0;
	}
	*pms = (now.tv_sec - epoch.tv_sec) * 1000;
	*pms += (now.tv_nsec - epoch.tv_nsec + 500000) / 1000000;
	return *pms;
}

/*************************************************************************
 * To compile:                                                           *
 * gcc -Wall -pedantic -std=c11 -g                                       *
 *     -o ../../apps/TCPtecho TCPtecho.c -L../../api -lconnectsock       *
 *                                                                       *
 * To run:                                                               *
 * ../../apps/TCPtecho localhost                                         *
 * ../../apps/TCPtecho ector arthur merlin                               *
 * ../../apps/TCPtecho -c 1000 sage                                      *
 *************************************************************************/
